/*
$FreeBSD$

The compilation of software known as FreeBSD GraphicBoot is distributed under
the following terms:

Copyright (C) 2005 Matthew Holder. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/
#ifndef _MAIN_C_
#define _MAIN_C_

#define MAX_LINE_SIZE		2048
#define CONNECT_TIMEOUT		30

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/Text.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <termios.h>
#include <fcntl.h>

XmString szGeneric;
Widget lblStatus, txtConsole;
static XtInputId input_id;
int nPTYFile, nTTYFile;

/* Function Prototypes */
int OpenConsole();
void CloseConsole();
int GetPTerm(int *pnTTY, int *pnPTY);

/* Callbacks */
void ProcessLine(/* Widget pWidget, int *pnFile, XtInputId *pID */);

/**
 * Startup
 */
int main(int argc, char *argv[])
{
	long nFailTime;
	Display *pDpy;
	
	XtAppContext xtAppContext;
	Arg args[20]; Cardinal n;
	
	Widget wndBoot, frmBoot;
	
	/*
	Since this program is designed to startup before an actual X
	server has been started, we will first try to start X, then
	attempt to connect to the server.
	*/
	system("/etc/Xboot/Xserver vt12 -quiet -config /etc/Xboot/xorg.conf -dpms -terminate -nolisten tcp > /dev/null&");

	nFailTime = time(NULL) + CONNECT_TIMEOUT;
	
	/* Only str to connect for so long then give up */		
	setenv("DISPLAY", ":0.0", TRUE);
	while(time(NULL) < nFailTime)
	{
		pDpy = XOpenDisplay(NULL);
		if (pDpy != NULL) break;
	}
	if (pDpy == NULL)
	{
		printf("Unable to use Graphical Boot\n");
		exit(0);
	}
	system("xsetroot -solid \"#0A246A\" -cursor_name watch");
	system("/usr/local/bin/xloadimage -onroot /etc/Xboot/Xsplash.jpg > /dev/null");
	system("xrdb /etc/Xboot/Xresources > /dev/null");
	
	wndBoot = XtVaOpenApplication
	(
		&xtAppContext, "FreeBSDGraphicBoot", (XrmOptionDescList)NULL, 0,
		&argc, argv, (String*)NULL, applicationShellWidgetClass, NULL
	);
	
	n = 0;
	XtSetArg(args[n], XmNmarginHeight, 7); n++;
	XtSetArg(args[n], XmNmarginWidth, 7); n++;
	XtSetArg(args[n], XmNverticalSpacing, 3); n++;
	XtSetArg(args[n], XmNhorizontalSpacing, 3); n++;
	frmBoot = XmCreateForm(wndBoot, "frmBoot", args, n);
	
	n = 0;
	szGeneric = XmStringCreateLocalized("Starting FreeBSD...");
	XtSetArg(args[n], XmNlabelString, szGeneric); n++;
	XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	lblStatus = XmCreateLabel(frmBoot, "lblStatus", args, n);
	XmStringFree(szGeneric);
	
	n = 0;
	XtSetArg(args[n], XmNeditMode, XmMULTI_LINE_EDIT); n++;
	XtSetArg(args[n], XmNeditable, FALSE); n++;
	XtSetArg(args[n], XmNrows, 15); n++;
	XtSetArg(args[n], XmNcolumns, 80); n++;
	XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
	XtSetArg(args[n], XmNtopWidget, lblStatus); n++;
	XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	txtConsole = XmCreateText(frmBoot, "txtConsole", args, n);
	
	XtManageChild(txtConsole);
	XtManageChild(lblStatus);
	XtManageChild(frmBoot);
	
	XtRealizeWidget(wndBoot);
	
	if (OpenConsole(txtConsole) == False) 
	{
		XCloseDisplay(pDpy);
		exit(0);
	}

	switch (fork())
	{
	case 0:
		break;
	case -1:
		printf("Fork Failure\n");
		exit(0);
		break;
	default:
		exit(0);
	}
	
	XtAppMainLoop(xtAppContext);
	
	XCloseDisplay(pDpy);
	exit(1);
	return 1;
}

/**
 * Attaches the console to a pseudo-terminal and opens that terminal
 * to read messages.
 *
 * @param (in) The widget to attach the pseudo-terminal to
 * @return True on success or False on error
 */
int OpenConsole()
{
	if (GetPTerm(&nTTYFile, &nPTYFile))
	{
		input_id = XtAppAddInput
		(
			XtWidgetToApplicationContext(txtConsole),
			nPTYFile, (XtPointer) XtInputReadMask,
			(XtInputCallbackProc)ProcessLine, txtConsole
		);
		return True;
	} else return False;
}

/**
 * Closes the virtual console.
 */
void CloseConsole()
{
	XtRemoveInput(input_id);
	close(nPTYFile);
	close(nTTYFile);
}

/**
 * Searches for an availible Pseudo-Terminal and opens the two devices files
 * for it.
 *
 * @param (out) The file descriptor of the Teletype Terminal file
 * @param (out) The file descriptor of the Pseudo Terminal file
 * @return True on success false on error
 */
int GetPTerm(int *pnTTY, int *pnPTY)
{
	char szTTYDev[11], szPTYDev[11], *szTerm;
	int nOn = 1;
	for (szTerm = "0123456789abcdef"; *szTerm != 0; szTerm++)
	{
		sprintf(szTTYDev, "/dev/ttyp%c", *szTerm);
		sprintf(szPTYDev, "/dev/ptyp%c", *szTerm);
		if
		(
			(*pnPTY = open(szPTYDev, O_RDWR)) >= 0 &&
			(*pnTTY = open(szTTYDev, O_RDWR)) >= 0 &&
			ioctl(*pnPTY, TIOCCONS, (char*)&nOn) != -1
		) return True;	
		
		if (*pnTTY >= 0) close(*pnTTY);
		if (*pnPTY >= 0) close(*pnPTY);
	}
	return False;
}

/**
 * Gets a line of input from a file as it is writen(received).
 */
void ProcessLine (Widget pWidget, int *pnFile, XtInputId *pID)
{
	char szBuffer[MAX_LINE_SIZE];
	char *sz;
	int nBytes;
	XmTextPosition nTextPos;
	
	nBytes = read(*pnFile, szBuffer, MAX_LINE_SIZE);
	if (nBytes != 0)
	{
		for (sz = szBuffer; *sz != 0; sz++) if (*sz == '\r') *sz = ' ';
		if (memcmp(szBuffer, "_XGB_STAT: ", 11) == 0)
		{	/* Change the status message */
			szGeneric = XmStringCreateLocalized(szBuffer + 11);
			XtVaSetValues
			(
				lblStatus,
				XmNlabelString, szGeneric,
				NULL
			);
			XmStringFree(szGeneric);
		
		} else if (memcmp(szBuffer, "_XGB_EGB", 8) == 0) {
			/* End the graphical boot */
			CloseConsole();
			exit(0);
		} else {
			/* Add the line to the text widget */
			szBuffer[nBytes] = '\0';
			nTextPos = XmTextGetLastPosition(pWidget);
			XmTextInsert(pWidget, nTextPos, szBuffer);
			nTextPos = XmTextGetLastPosition(pWidget);
			XtVaSetValues
			(
				pWidget,
				XmNcursorPosition, nTextPos,
				NULL
			);
			XmTextShowPosition(pWidget, nTextPos);
		}			
		fflush(stdout);
	} else XtRemoveInput(*pID);
}

#endif /*_MAIN_C_*/
